@@ -1,6 +1,8 @@
 # -*- coding: utf-8 -*-
+"""Back-port compiler for Python 3.8 positional-only parameter syntax."""
 
 import argparse
+import contextlib
 import glob
 import locale
 import os
@@ -31,7 +33,7 @@ finally:    # alias and aftermath
     del multiprocessing
 
 # version string
-__version__ = '0.4.1'
+__version__ = '0.4.2.post1'
 
 # from configparser
 BOOLEAN_STATES = {'1': True, '0': False,
@@ -73,6 +75,38 @@ def predicate(filename):  # pragma: no cover
 
 tbtrim.set_trim_rule(predicate, strict=True, target=(ConvertError, EnvironError))
 
+###############################################################################
+# UUID 4 generator wrapper
+
+
+class UUID4Generator:
+    """UUID 4 generator wrapper to prevent UUID collisions."""
+
+    def __init__(self, dash=True):
+        """Constructor of UUID 4 generator wrapper.
+        Args:
+         - `dash` -- `bool`, whether the generated UUID string has dashes or not
+        """
+        self.used_uuids = set()
+        self.dash = dash
+
+    def gen(self):
+        """Generate a new UUID 4 string that is guaranteed not to collide with used UUIDs.
+        Returns:
+         - `str` -- a new UUID 4 string
+        """
+        while True:
+            nuid = uuid.uuid4()
+            nuid = str(nuid) if self.dash else nuid.hex
+            if nuid not in self.used_uuids:
+                break
+        self.used_uuids.add(nuid)
+        return nuid
+
+
+# Initialise a UUID4Generator for filename UUIDs.
+uuid_gen = UUID4Generator(dash=True)
+
 ###############################################################################
 # Positional-only decorator
 
@@ -98,13 +132,12 @@ exec(os.linesep.join(_decorator) % dict(decorator='decorator', tabsize='\t'.expa
 # Main convertion implementation
 
 
-def parse(string, source, error_recovery=False):
+def parse(string, source):
     """Parse source string.
 
     Args:
      - `string` -- `str`, context to be converted
      - `source` -- `str`, source of the context
-     - `error_recovery` -- `bool`, see `parso.Grammar.parse`
 
     Envs:
      - `POSEUR_VERSION` -- convert against Python version (same as `--python` option in CLI)
@@ -113,16 +146,24 @@ def parse(string, source, error_recovery=False):
      - `parso.python.tree.Module` -- parso AST
 
     Raises:
-     - `ConvertError` -- when `parso.ParserSyntaxError` raised
+     - `ConvertError` -- when source code contains syntax errors
 
     """
-    try:
-        return parso.parse(string, error_recovery=error_recovery,
-                           version=os.getenv('POSEUR_VERSION', POSEUR_VERSION[-1]))
-    except parso.ParserSyntaxError as error:
-        message = '%s: <%s: %r> from %s' % (error.message, error.error_leaf.token_type,
-                                            error.error_leaf.value, source)
-        raise ConvertError(message).with_traceback(error.__traceback__) from None
+    grammar = parso.load_grammar(version=os.getenv('POSEUR_VERSION', POSEUR_VERSION[-1]))
+    module = grammar.parse(string, error_recovery=True)
+
+    # suppress parso debug outputs
+    # see https://github.com/davidhalter/parso/commit/ee2995c1108028cd8c6d56df3a00170c93dc9ce6
+    # TODO: remove this redirect after a new version of parso is released
+    with open(os.devnull, 'w') as nul_file:
+        with contextlib.redirect_stdout(nul_file):
+            errors = grammar.iter_errors(module)
+
+    if errors:
+        error_messages = '\n'.join('[L%dC%d] %s' % (*error.start_pos, error.message) for error in errors)
+        raise ConvertError('source file %r contains following syntax errors:\n' % source + error_messages)
+
+    return module
 
 
 def decorate_lambdef(parameters, lambdef):
@@ -769,14 +810,14 @@ def poseur(filename):
     encoding = os.getenv('POSEUR_ENCODING', LOCALE_ENCODING)
 
     # file content
-    with open(filename, 'r', encoding=encoding) as file:
+    with open(filename, 'r', encoding=encoding, newline='') as file:
         text = file.read()
 
     # do the dirty things
     text = convert(text, filename)
 
     # dump back to the file
-    with open(filename, 'w', encoding=encoding) as file:
+    with open(filename, 'w', encoding=encoding, newline='') as file:
         file.write(text)
 
 
@@ -871,7 +912,7 @@ def rename(path, root):
 
     """
     stem, ext = os.path.splitext(path)
-    name = '%s-%s%s' % (stem, uuid.uuid4(), ext)
+    name = '%s-%s%s' % (stem, uuid_gen.gen(), ext)
     return os.path.join(root, name)
 
 
